@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2021 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -27,4 +27,4 @@ Batch System
 from __future__ import unicode_literals, absolute_import
 
 from .handlers import BatchHandler, get_batch_handler
-from .util import consume_batch_id
+from .util import consume_batch_id, batch_id_str
