@@ -26,6 +26,9 @@ Database Stuff
 
 from __future__ import unicode_literals, absolute_import
 
+import os
+import sys
+
 try:
     import sqlalchemy
 except ImportError:
@@ -223,3 +226,15 @@ class ConfigExtension(BaseExtension):
                               'trainwreck import-trainwreck')
             config.setdefault('rattail.importing', 'to_trainwreck.from_trainwreck.import.legacy_handler_setting',
                               'trainwreck.importing, trainwreck.handler')
+
+            # nb. cannot fetch DB settings during init
+            appdir = config.appdir(require=False, usedb=False)
+            if not appdir:
+                appdir = os.path.join(sys.prefix, 'app')
+
+            poser = config.get('rattail', 'poser', usedb=False,
+                               default=os.path.join(appdir, 'poser'))
+
+            # add poser to path if it exists
+            if os.path.isdir(poser) and poser not in sys.path:
+                sys.path.append(poser)
