@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2021 Lance Edgar
+#  Copyright © 2010-2022 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -88,6 +88,15 @@ class tailbone_buefy_version(Setting):
     name = 'buefy_version'
 
 
+class tailbone_grid_default_pagesize(Setting):
+    """
+    Default page size for grids.
+    """
+    namespace = 'tailbone'
+    name = 'grid.default_pagesize'
+    data_type = int
+
+
 class tailbone_sticky_headers(Setting):
     """
     Whether table/grid headers should be "sticky" for *ALL* grids.
